/*
 Ŀ
                      RS232 communications                         
 Ĵ
                                                                   
    Version  :  1.00                                               
    Filename :  RS232.CPP                                          
                                                                   
    Author   :  Mats Dal                                           
    Email    :  d92-mda@nada.kth.se                                
                                                                   
    Ver/Date :  1.00 960727 (YYMMDD)                               
                                                                   
    Descr    :  RS232 procedures.                                  
                                                                   
    News     :  1.00 First release. RTS/CTS hardware flowctrl.     
                     Supports 16550, multiopen ports etc.          
                     No copyright, freeware.                       
                                                                   
 
*/

#include <dos.h>
#include <conio.h>
#include <stdio.h>
#include <stdlib.h>

#include "RS232.H"

#define ON 	1
#define OFF 0

#define NPORTS 4                        // Number of Comports.

/* Modem CTRL */
#define RTS  2
#define DTR  1
#define OUT2 8

/* ModemStatus */
#define DCD 128
#define RI 	64
#define DSR 32
#define CTS 16

/* Identify which interrupts that has occured. */
#define RECIEVE 4
#define THR 2
#define LINE 6
#define STATUS 0

/* Add these to UARTbase[port] */
#define MSR 6
#define LSR 5
#define MCR 4
#define LCR 3
#define IIR 2
#define IER 1

// Procedures.
void SetModemCTRL(byte ctrl, int port);
void interrupt RS232_1hand(void);
void interrupt RS232_2hand(void);
void interrupt RS232_3hand(void);
void interrupt RS232_4hand(void);
void interrupt (__far *OldHandler[NPORTS])(void);
void Set8259(int sw, int port);
int DetectUART(int port);

// Data.
static byte PortUsed[] = {0, 0, 0, 0};
byte Msk8259[] = {0xEF, 0xF7, 0xEF, 0xF7};
word Inter[] = {0x0C, 0x0B, 0x0C, 0x0B};
word UARTbase[] = {0x3F8, 0x2F8, 0x3E8, 0x2E8};
word RXfifoSize[] = {4096, 4096, 4096, 4096};
word TXfifoSize[] = {4096, 4096, 4096, 4096};
byte bitxfifo[NPORTS];

volatile byte *RXfifo[NPORTS], *TXfifo[NPORTS];
volatile word RXtail[NPORTS], RXhead[NPORTS];
volatile word TXtail[NPORTS], TXhead[NPORTS];
volatile word RXfilled[NPORTS];
volatile byte TXon[NPORTS], TXdisabled[NPORTS], RTSon[NPORTS];

void interrupt (__far *handlers[])(void) = {RS232_1hand, RS232_2hand,
                                            RS232_3hand, RS232_4hand};

/*
 ͻ
                                                                           
  Name: SetBaud()                                                          
  Desc: Set baudrate to UART.                                              
  Inpt: baud = 115200, 57600, 38400, 19200, 9600, 4800, 2400, 1200 or 300. 
  Outp: None.                                                              
                                                                           
 ͼ
*/
void SetBaud(int baud, int port)
{
  word d;

  d = 384 / (baud / 300);

  _disable();

  // DLAB ON
  outp(UARTbase[port]+LCR, inp(UARTbase[port]+LCR) | 0x80);

  outp(UARTbase[port], d & 0x00FF);
  outp(UARTbase[port]+1, d >> 16);

  // DLAB OFF
  outp(UARTbase[port]+LCR, inp(UARTbase[port]+LCR) & 0x7F);

  _enable();
}

void ClearFIFOS(int port)
{
  _disable();

  // enable RECIEVE, MODEMSTATUS and LINESTATUS interrupts.
  outp(UARTbase[port] + IER, 1 | 4 | 8);
  TXdisabled[port] = 1;

  // read modemstatus register.
  TXon[port] = (inp(UARTbase[port]+MSR) & CTS) ? 1 : 0;
  
  RXfilled[port] = RXfifoSize[port];     // Number of bytes not used in buffer.
  RXtail[port] = 0;
  RXhead[port] = 0;
  TXtail[port] = 0;
  TXhead[port] = 0;

  _enable();
}

/*
 ͻ
                                                                           
  Name: m_gets()                                                           
  Desc: Reads multiple chars at once.                                      
  Inpt: Ptr to storage and length of array.                                
  Outp: From FIFOs to *s.                                                  
                                                                           
 ͼ
*/
void m_gets(char *s, int length, int port)
{
  int i;
  byte *RXbuf = RXfifo[port];

  if (!PortUsed[port])                // Check if port is open!
    return;

  for (i = 0; i < length; i++)
  {
    while ((volatile word) RXfilled[port] == RXfifoSize[port]);   // No data in Recieve FIFO.

    *s = RXbuf[RXtail[port]];
    s++;
    RXtail[port]++;
    if (RXtail[port] == RXfifoSize[port])
      RXtail[port] = 0;
    RXfilled[port]++;
    if (RTSon[port] == 0 && RXfilled[port] >= 33)
    {
      SetModemCTRL(RTS | DTR, port); RTSon[port] = 1;
    }
  }
}

int m_getc(int port)
{
  int c;

  if (!PortUsed[port])                // Check if port is open!
    return -1;

  while ((volatile word) RXfilled[port] == RXfifoSize[port]);     // No data in Recieve FIFO.

  c = (int) RXfifo[port][RXtail[port]];

  RXtail[port]++;
  if (RXtail[port] == RXfifoSize[port])
    RXtail[port] = 0;
  RXfilled[port]++;
  if (RTSon[port] == 0 && RXfilled[port] >= 33)
  {
    SetModemCTRL(RTS | DTR, port); RTSon[port] = 1;
  }

	return c;
}

int m_input(int port)
{
  return !(RXfilled[port] == RXfifoSize[port]);
}

/*
 ͻ
                                                                           
  Name: m_inFilled()                                                       
  Desc: Calculates how many bytes that are available in input buffer.      
  Inpt: None.                                                              
  Outp: n bytes that are read into input buffer.                           
                                                                           
 ͼ
*/
int m_inFilled(int port)
{
  return (RXfifoSize[port] - RXfilled[port]);
}

void m_putc(int c, int port)
{
  word head;

  if (!PortUsed[port])                // Check if port is open!
    return;

  head = TXhead[port];

	head++;
  if (head == TXfifoSize[port])
		head = 0;

  for (;;)
  {
    if (head != TXtail[port])         // Wait while buffers are full.
      break;
  }

  TXfifo[port][TXhead[port]] = (byte) c;
  TXhead[port] = head;

  _disable();
  if (TXon[port] && TXdisabled[port] && TXtail[port] != TXhead[port])
  {
    outp(UARTbase[port] + IER, 1 | 2 | 4 | 8); // Enable transmit interrupts.
    TXdisabled[port] = 0;
  }
  _enable();
}

/*
 ͻ
                                                                           
  Name: OkToOutput()                                                       
  Desc: Is there room for nBytes characters in TX fifo?                    
  Inpt: nBytes to check.                                                   
  Outp: 1 == OK, 0 == NOT OK.                                              
                                                                           
 ͼ
*/
int OkToOutput(int nBytes, int port)
{
  word Head, Tail;
  byte HeadLesser;

  Tail = TXtail[port];
  Head = TXhead[port];

  HeadLesser = 0;
  if (Head < Tail)
    HeadLesser = 1;

  Head += nBytes;
  if (Head >= TXfifoSize[port])
  {
    Head -= TXfifoSize[port];
    if (HeadLesser && Head <= Tail || !HeadLesser && Head >= Tail)
      return 0;
  }
  else
  {
    if (HeadLesser && Head >= Tail)
      return 0;
  }

  return 1;
}

/*
 ͻ
                                                                           
  Name: m_puts()                                                           
  Desc: Outputs length bytes to TX fifo.                                   
  Inpt: buffer, length.                                                    
  Outp: None.                                                              
                                                                           
 ͼ
*/
void m_puts(char *str, int length, int port)
{
  word Head;
  int tail;
  byte *TXbuf = TXfifo[port];

  if (!PortUsed[port])                // Check if port is open!
    return;

  Head = TXhead[port];

  while (length--)
  {
    TXbuf[Head] = *str;
    str++;
    Head++;
    if (Head == TXfifoSize[port])
      Head = 0;

    for (;;)
    {
      _disable();
      tail = TXtail[port];
      _enable();
      if (tail != Head)
        break;
    }
  }

  TXhead[port] = Head;

  _disable();
  if (TXon[port] && TXdisabled[port] && TXtail[port] != TXhead[port])
  {
    outp(UARTbase[port] + IER, 1 | 2 | 4 | 8); // Enable transmit interrupts.
    TXdisabled[port] = 0;
  }
  _enable();
}


/*
 ͻ
                                                                           
  Name: SetModemCTRL()                                                     
  Desc: Sets modem control register.                                       
  Inpt: ctrl.                                                              
  Outp: None.                                                              
                                                                           
 ͼ
*/
void SetModemCTRL(byte ctrl, int port)
{
  outp(UARTbase[port] + MCR, OUT2 | ctrl);
}

/*
 ͻ
                                                                           
  Name: OpenRS232()                                                        
  Desc: Open communicationports.                              							
  Inpt: Port number.                                                       
  Outp: 1 = UARTtype, 0 = Error.                                           
                                                                           
 ͼ
*/
int OpenRS232(int port, int baud)
{
  int i;

  if (PortUsed[port])
    return 0;             // Return error if port already is open.

  i = DetectUART(port);

  if (i == 4 || i == 3)   // Only enable FIFO's if 16550.
	{
    printf("16550 detected\n");
    bitxfifo[port] = 16;
    outp(UARTbase[port]+2, 0xC7);    // Enable 16550A fifo's.
	}
  else
	{
    bitxfifo[port] = 1;

    if (i == 0)
      return 0;
	}

  SetBaud(baud, port);
  outp(UARTbase[port] + LCR, 3);      // n,8,1 as default.

  if ((RXfifo[port] = (byte *) malloc(RXfifoSize[port])) == NULL)
    return 0;
  if ((TXfifo[port] = (byte *) malloc(TXfifoSize[port])) == NULL)
    return 0;

	// Install our own new interrupt.
  OldHandler[port] = _dos_getvect(Inter[port]);
  _dos_setvect(Inter[port], handlers[port]);

  // Reset our FIFOs.
  ClearFIFOS(port);

  // enable RECIEVE, MODEMSTATUS and LINESTATUS interrupts.
  outp(UARTbase[port] + IER, 1 | 4 | 8);
	// clear reciever.
  inp(UARTbase[port]);
	// clear line status.
  inp(UARTbase[port]+LSR);
  // clear modemstatus reg.
  inp(UARTbase[port]+MSR);

	// free interrupt in the interrupt controller unit.
	Set8259(ON, port);

	// Enable interrupts from the UART.
  outp(UARTbase[port] + MCR, OUT2 | RTS | DTR);
  RTSon[port] = 1;

  // Activate port.
  PortUsed[port] = 1;

  return i;
}

/*
 ͻ
                                                                           
  Name: CloseRS232()                                                       
  Desc: Close active communicationport.                       							
  Inpt: port. (0-3)                                                        
  Outp: None.                                                              
                                                                           
 ͼ
*/
void CloseRS232(int port)
{
  if (!PortUsed[port])                // Check if port is open!
		return;

	// Turn off our communication interrupts.
  Set8259(OFF, port);

	// Disconnect the UART from the int line.
  outp(UARTbase[port] + MCR, 0);

	// Disable UART ints.
  outp(UARTbase[port] + IER, 0);

  // disable 16550A fifos.
  if (bitxfifo[port] > 1)
    outp(UARTbase[port] + 2, 0x0);

	// Restore old interrupt handler.
  _dos_setvect(Inter[port], OldHandler[port]);

  free((void *) RXfifo[port]);
  free((void *) TXfifo[port]);

	// Deactive port.
  PortUsed[port] = 0;
}

/*
 ͻ
                                                                           
  Name: Set8259()                                                          
  Desc: This turns the communications interrupts on or off.  							
  Inpt: sw, port                                                           
  Outp: None.                                                              
                                                                           
 ͼ
*/
void Set8259(int sw, int port)
{
  // stop interrupts while we do this.
  _disable();
	switch (sw)
	{
    case ON: outp(0x21, inp(0x21) & Msk8259[port]); break;
    case OFF: outp(0x21, inp(0x21) | (0xFF ^ Msk8259[port])); break;
	}
  _enable();
}

/*
 ͻ
                                                                           
  Name: DetectUART()                                                       
  Desc: This function returns 0 if no UART is installed.                   
  Inpt: port                                                               
  Outp: 1: 8250, 2: 16450 or 8250 with scratch reg., 3: 16550, 4: 16550A   
                                                                           
 ͼ
*/
int DetectUART(int port)
{
   int x, olddata;
	 unsigned baseaddr;

   baseaddr = UARTbase[port];

	 // check if a UART is present anyway
	 olddata=inp(baseaddr+4);
	 outp(baseaddr+4,0x10);
	 if ((inp(baseaddr+6)&0xf0)) return 0;
	 outp(baseaddr+4,0x1f);
	 if ((inp(baseaddr+6)&0xf0)!=0xf0) return 0;
	 outp(baseaddr+4,olddata);
	 // next thing to do is look for the scratch register
	 olddata=inp(baseaddr+7);
	 outp(baseaddr+7,0x55);
	 if (inp(baseaddr+7)!=0x55) return 1;
	 outp(baseaddr+7,0xAA);
	 if (inp(baseaddr+7)!=0xAA) return 1;
   outp(baseaddr+7,olddata);

   // then check if there's a FIFO
	 outp(baseaddr+2,1);
	 x=inp(baseaddr+2);
	 // some old-fashioned software relies on this!
	 outp(baseaddr+2,0x0);
	 if ((x&0x80)==0) return 2;
	 if ((x&0x40)==0) return 3;

	 return 4;
}
